class PullOrdersSimultaneousLimitExceeded(Exception):

    def __init__(
        self,               # type: PullOrdersSimultaneousLimitExceeded
    ):
        # type: (...) -> None
        super(PullOrdersSimultaneousLimitExceeded, self).__init__()
