from typing import Optional


class Fees(object):

    def __init__(
        self,                           # type: Fees
        code,                           # type: str
        included_in_price=False,        # type: bool
        calculated_amount=None,         # type: Optional[float]
    ):
        # type: (...) -> None

        self.code = code
        self.included_in_price = included_in_price
        self.calculated_amount = calculated_amount
