class PosControlErrorCodes:
    SE_UNKNOWNUSR = 5000
    SE_INVALIDPASSWD = 5001
    SE_DAYNOTCLOSED = 5002
    SE_DAYNOTOPENED = 5003
    SE_OPERLOGGEDIN = 5004
    SE_OPERNOTLOGGEDIN = 5005
    SE_ORDEROPENED = 5006
    SE_OPEROPENED = 5007
