from typing import Optional


class User(object):
    def __init__(
        self,       # type: User
        id,         # type: int
        name,       # type: Optional[str]
        level,      # type: Optional[int]
    ):
        self.id = id
        self.name = name
        self.level = level
